# -*- coding: utf-8 -*-
# Minio Python Library for Amazon S3 compatible cloud storage, (C) 2015 Minio, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytz

from xml.etree import cElementTree
from xml.etree.cElementTree import ParseError
from datetime import datetime

from .acl import Acl
from .compat import urldecode
from .error import ResponseError
from .definitions import (Object, Bucket, IncompleteUpload, UploadPart)

def parse_list_buckets(data):
    root = cElementTree.fromstring(data)
    bucket_list = []
    for buckets in root:
        if buckets.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}Buckets':
            for bucket in buckets:
                name = None
                creation_date = None
                for attribute in bucket:
                    if attribute.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}Name':
                        name = attribute.text
                    if attribute.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}CreationDate':
                        creation_date = _parse_date(attribute.text)
                bucket_list.append(Bucket(name, creation_date))
    return bucket_list


def parse_acl(data):
    root = cElementTree.fromstring(data)

    public_read = False
    public_write = False
    authenticated_read = False
    authenticated_write = False

    for acls in root:
        if acls.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}AccessControlList':
            for grant in acls:
                user_uri = None
                permission = None
                for grant_property in grant:
                    if grant_property.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}Grantee':
                        for grantee in grant_property:
                            if grantee.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}URI':
                                user_uri = grantee.text
                    if grant_property.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}Permission':
                        permission = grant_property.text
                if user_uri == 'http://acs.amazonaws.com/groups/global/AllUsers' and permission == 'WRITE':
                    public_write = True
                if user_uri == 'http://acs.amazonaws.com/groups/global/AllUsers' and permission == 'READ':
                    public_read = True
                if user_uri == 'http://acs.amazonaws.com/groups/global/AuthenticatedUsers' and permission == 'READ':
                    authenticated_read = True
                if user_uri == 'http://acs.amazonaws.com/groups/global/AuthenticatedUsers' and permission == 'WRITE':
                    authenticated_write = True

    if public_read is True and public_write is True:
        return Acl.public_read_write()
    if public_read is True and public_write is False:
        return Acl.public_read()
    if authenticated_read is True and authenticated_write is False:
        return Acl.authenticated_read()
    return Acl.private()


def parse_list_objects(data, bucket):
    root = cElementTree.fromstring(data)

    is_truncated = False
    objects = []
    marker = None
    last_key = None
    for contents in root:
        if contents.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}IsTruncated':
            is_truncated = contents.text.lower() == 'true'
        if contents.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}NextMarker':
            if contents.text is not None:
                marker = urldecode(contents.text)
        if contents.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}Contents':
            key = None
            last_modified = None
            etag = None
            size = None
            for content in contents:
                if content.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}Key':
                    key = urldecode(content.text)
                    last_key = key
                if content.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}LastModified':
                    last_modified = _parse_date(content.text)
                if content.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}ETag':
                    etag = content.text
                    etag = etag.replace('"', '')
                if content.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}Size':
                    size = content.text
            objects.append(Object(bucket, key, last_modified, etag, size, content_type=None))
        if contents.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}CommonPrefixes':
            for content in contents:
                if content.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}Prefix':
                    key = urldecode(content.text)
                # noinspection PyUnboundLocalVariable
                objects.append(Object(bucket, key, None, '', 0, content_type=None, is_dir=True))

    if is_truncated and marker is None:
        marker = last_key

    return objects, is_truncated, marker


def parse_incomplete_uploads(data, bucket):
    root = cElementTree.fromstring(data)

    is_truncated = False
    uploads = []
    key_marker = None
    upload_id_marker = None
    for contents in root:
        if contents.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}IsTruncated':
            is_truncated = contents.text.lower() == 'true'
        if contents.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}NextKeyMarker':
            if contents.text is not None:
                key_marker = urldecode(contents.text)
        if contents.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}NextUploadIdMarker':
            upload_id_marker = contents.text
        if contents.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}Upload':
            key = None
            upload_id = None
            for content in contents:
                if content.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}Key':
                    key = urldecode(content.text)
                if content.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}UploadId':
                    upload_id = content.text
            uploads.append(IncompleteUpload(bucket, key, upload_id))

    return uploads, is_truncated, key_marker, upload_id_marker


def parse_uploaded_parts(data, bucket, key, upload_id):
    root = cElementTree.fromstring(data)

    is_truncated = False
    parts = []
    part_marker = None
    for contents in root:
        if contents.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}IsTruncated':
            is_truncated = contents.text.lower() == 'true'
        if contents.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}NextPartNumberMarker':
            part_marker = contents.text
        if contents.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}Part':
            part_number = None
            etag = None
            last_modified = None
            size = None
            for content in contents:
                if content.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}PartNumber':
                    part_number = int(content.text)
                if content.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}ETag':
                    etag = content.text
                    etag = etag.replace('"', '')
                if content.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}LastModified':
                    last_modified = _parse_date(content.text)
                if content.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}Size':
                    size = content.text
            parts.append(UploadPart(bucket, key, upload_id, part_number, etag,
                                    last_modified, size))
    return parts, is_truncated, part_marker


def parse_new_multipart_upload(data):
    root = cElementTree.fromstring(data)

    for contents in root:
        if contents.tag == '{http://s3.amazonaws.com/doc/2006-03-01/}UploadId':
            return contents.text

    raise ParseError('uploadId')

def parse_error(response, resource=None):
    if len(response.data) == 0:
        amz_request_id = ''
        amz_host_id = ''
        if 'x-amz-request-id' in response.headers:
            amz_request_id = response.headers['x-amz-request-id']
        if 'x-amz-id-2' in response.headers:
            amz_host_id = response.headers['x-amz-id-2']
        if response.status == 405 or response.status == 501:
            raise ResponseError('MethodNotAllowedException', response.reason,
                                amz_request_id, amz_host_id, resource, None)
        if response.status == 404:
            raise ResponseError('ObjectNotFoundException', response.reason,
                                amz_request_id, amz_host_id, resource, None)
        if response.status == 403:
            raise ResponseError('AccessDeniedException', response.reason,
                                amz_request_id, amz_host_id, resource, None)
        if response.status == 400:
            raise ResponseError('BadRequestException', response.reason,
                                amz_request_id, amz_host_id, resource, None)
        if response.status == 301 or response.status == 307:
            raise ResponseError('Redirect', response.reason,
                                amz_request_id, amz_host_id, resource, None)
        raise ResponseError('UnknownException', response.reason,
                            amz_request_id, amz_host_id, resource, None)

    code = None
    message = None
    request_id = None
    host_id = None
    resource = None

    root = cElementTree.fromstring(response.data)
    for attribute in root:
        if attribute.tag == 'Code':
            code = attribute.text
        if attribute.tag == 'Message':
            message = attribute.text
        if attribute.tag == 'RequestId':
            request_id = attribute.text
        if attribute.tag == 'HostId':
            host_id = attribute.text
        if attribute.tag == 'Resource':
            resource = attribute.text

    raise ResponseError(code, message,
                        request_id, host_id, resource,
                        response.data)

def _parse_date(date_string):
    parsed_date = datetime.strptime(date_string, '%Y-%m-%dT%H:%M:%S.%fZ')
    localized_time = pytz.utc.localize(parsed_date)
    return localized_time



# -*- coding: utf-8 -*-
# Minio Python Library for Amazon S3 compatible cloud storage, (C) 2015 Minio, Inc.
#
# Licensed under the Apache License, Version 2.0 (the 'License');
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an 'AS IS' BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class Acl(object):
    @staticmethod
    def public_read_write():
        """ Public Read Write ACL """
        return 'public-read-write'

    @staticmethod
    def public_read():
        """ Public Read ACL"""
        return 'public-read'

    @staticmethod
    def authenticated_read():
        """ Authenticated Users Read """
        return 'authenticated-read'

    @staticmethod
    def private():
        """ Owner Read Write ACL"""
        return 'private'


def is_valid_acl(acl):
    if acl == Acl.public_read_write() or acl == Acl.public_read() \
            or acl == Acl.authenticated_read() or acl == Acl.private():
        return
    raise ValueError()



# -*- coding: utf-8 -*-
# Minio Python Library for Amazon S3 compatible cloud storage, (C) 2015 Minio, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class Bucket(object):
    def __init__(self, name, created):
        self.name = name
        self.creation_date = created

    def __str__(self):
        return '<Bucket: {0} {1}>'.format(self.name, self.creation_date)


class Object(object):
    def __init__(self, bucket, key, last_modified, etag, size,
                 content_type=None, is_dir=False):
        self.bucket = bucket
        self.key = key
        self.last_modified = last_modified
        self.etag = etag
        self.size = size
        self.content_type = content_type
        self.is_dir = is_dir

    def __str__(self):
        string_format = '<Object: bucket: {0} key: {1} last_modified: {2}' \
                        ' etag: {3} size: {4} content_type: {5}, is_dir: {6}>'
        return string_format.format(self.bucket, self.key, self.last_modified,
                                    self.etag, self.size, self.content_type,
                                    self.is_dir)

class IncompleteUpload(object):
    def __init__(self, bucket, key, upload_id):
        self.bucket = bucket
        self.key = key
        self.upload_id = upload_id

    def __str__(self):
        string_format = '<IncompleteUpload: bucket: {0} key: {1}' \
                        ' upload_id: {2}>'
        return string_format.format(self.bucket, self.key, self.upload_id)

class UploadPart(object):
    def __init__(self, bucket, key, upload_id, part_number, etag,
                 last_modified, size):
        self.bucket = bucket
        self.key = key
        self.upload_id = upload_id
        self.part_number = part_number
        self.etag = etag
        self.last_modified = last_modified
        self.size = size

    def __str__(self):
        string_format = '<UploadPart: bucket: {0} key: {1} upload_id: {2}' \
                        ' part_number: {3} etag: {4} last_modified: {5}' \
                        ' size: {6}>'
        return string_format.format(self.bucket,
                                    self.key,
                                    self.upload_id,
                                    self.part_number,
                                    self.etag,
                                    self.last_modified,
                                    self.size)



# -*- coding: utf-8 -*-
# Minio Python Library for Amazon S3 compatible cloud storage, (C) 2015 Minio, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from io import BytesIO
from xml.etree import ElementTree

__author__ = 'minio'


def bucket_constraint(region):
    root = ElementTree.Element('CreateBucketConfiguration', {'xmlns': 'http://s3.amazonaws.com/doc/2006-03-01/'})
    location_constraint = ElementTree.SubElement(root, 'LocationConstraint')
    location_constraint.text = region
    data = BytesIO()
    ElementTree.ElementTree(root).write(data, encoding=None, xml_declaration=False)
    return data.getvalue()


def get_complete_multipart_upload(etags):
    root = ElementTree.Element('CompleteMultipartUpload', {'xmlns': 'http://s3.amazonaws.com/doc/2006-03-01/'})

    for i in range(0, len(etags)):
        part = ElementTree.SubElement(root, 'Part')
        part_number = ElementTree.SubElement(part, 'PartNumber')
        part_number.text = str(i + 1)
        etag = ElementTree.SubElement(part, 'ETag')
        etag.text = etags[i]
        data = BytesIO()
        ElementTree.ElementTree(root).write(data, encoding=None, xml_declaration=False)
    return data.getvalue()



# -*- coding: utf-8 -*-
# Minio Python Library for Amazon S3 compatible cloud storage, (C) 2015 Minio, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
import hmac
import binascii

from datetime import datetime
from .compat import urlsplit, strtype
from .helpers import get_region

def sign_v4(method, url, headers=None, access_key=None, secret_key=None,
            content_hash=None):
    if not access_key or not secret_key:
        return headers

    if headers is None:
        headers = {}

    parsed_url = urlsplit(url)
    content_hash_hex = 'e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855'
    if content_hash is not None:
        content_hash_hex = binascii.hexlify(content_hash).decode('utf-8')

    host = parsed_url.netloc
    headers['host'] = host
    headers['x-amz-date'] = datetime.utcnow().strftime("%Y%m%dT%H%M%SZ")
    headers['x-amz-content-sha256'] = content_hash_hex

    headers_to_sign = dict(headers)

    # Excerpts from @lsegal - https://github.com/aws/aws-sdk-js/issues/659#issuecomment-120477258
    #
    #  User-Agent:
    #
    #      This is ignored from signing because signing this causes problems with generating pre-signed URLs
    #      (that are executed by other agents) or when customers pass requests through proxies, which may
    #      modify the user-agent.
    #
    #  Content-Length:
    #
    #      This is ignored from signing because generating a pre-signed URL should not provide a content-length
    #      constraint, specifically when vending a S3 pre-signed PUT URL. The corollary to this is that when
    #      sending regular requests (non-pre-signed), the signature contains a checksum of the body, which
    #      implicitly validates the payload length (since changing the number of bytes would change the checksum)
    #      and therefore this header is not valuable in the signature.
    #
    #  Content-Type:
    #
    #      Signing this header causes quite a number of problems in browser environments, where browsers
    #      like to modify and normalize the content-type header in different ways. There is more information
    #      on this in https://github.com/aws/aws-sdk-js/issues/244. Avoiding this field simplifies logic
    #      and reduces the possibility of future bugs
    #
    #  Authorization:
    #
    #      Is skipped for obvious reasons

    ignored_headers = ['Authorization', 'Content-Length', 'Content-Type',
                       'User-Agent']

    for ignored_header in ignored_headers:
        if ignored_header in headers_to_sign:
            del headers_to_sign[ignored_header]

    canonical_request, signed_headers = generate_canonical_request(method,
                                                                   parsed_url,
                                                                   headers_to_sign,
                                                                   content_hash_hex)

    region = get_region(parsed_url.hostname)

    date = datetime.utcnow()

    canonical_request_hasher = hashlib.sha256()
    canonical_request_hasher.update(canonical_request.encode('utf-8'))
    canonical_request_sha256 = canonical_request_hasher.hexdigest()

    string_to_sign = generate_string_to_sign(date, region,
                                             canonical_request_sha256)
    signing_key = generate_signing_key(date, region, secret_key)
    signed_request = hmac.new(signing_key, string_to_sign.encode('utf-8'),
                              hashlib.sha256).hexdigest()

    authorization_header = generate_authorization_header(access_key, date, region,
                                                         signed_headers,
                                                         signed_request)

    headers['authorization'] = authorization_header

    return headers


def generate_canonical_request(method, parsed_url, headers, content_hash_hex):
    content_hash_hex = str(content_hash_hex)
    lines = [method, parsed_url.path]

    split_query = parsed_url.query.split('&')
    split_query.sort()
    for i in range(0, len(split_query)):
        if len(split_query[i]) > 0 and '=' not in split_query[i]:
            split_query[i] += '='

    query = '&'.join(split_query)
    lines.append(query)

    signed_headers = []
    header_lines = []
    for header in headers:
        value = headers[header]
        if isinstance(value, strtype):
            value = value.strip()
        header = header.lower().strip()
        signed_headers.append(header)
        header_lines.append(header.lower().strip() + ':' + str(value))
    signed_headers.sort()
    header_lines.sort()
    lines = lines + header_lines

    lines.append('')

    lines.append(';'.join(signed_headers))
    lines.append(str(content_hash_hex))

    return '\n'.join(lines), signed_headers


def generate_string_to_sign(date, region, request_hash):
    formatted_date_time = date.strftime("%Y%m%dT%H%M%SZ")
    formatted_date = date.strftime("%Y%m%d")

    scope = '/'.join([formatted_date,
                      region,
                      's3',
                      'aws4_request'])

    return '\n'.join(['AWS4-HMAC-SHA256',
                      formatted_date_time,
                      scope,
                      request_hash])


def generate_signing_key(date, region, secret):
    formatted_date = date.strftime("%Y%m%d")

    key1_string = 'AWS4' + secret
    key1 = key1_string.encode('utf-8')
    key2 = hmac.new(key1, formatted_date.encode('utf-8'),
                    hashlib.sha256).digest()
    key3 = hmac.new(key2, region.encode('utf-8'), hashlib.sha256).digest()
    key4 = hmac.new(key3, 's3'.encode('utf-8'), hashlib.sha256).digest()

    return hmac.new(key4, 'aws4_request'.encode('utf-8'),
                    hashlib.sha256).digest()


def generate_authorization_header(access_key, date, region, signed_headers,
                                  signed_request):
    formatted_date = date.strftime("%Y%m%d")
    signed_headers_string = ';'.join(signed_headers)
    auth_header = "AWS4-HMAC-SHA256 Credential=" + access_key + "/" + \
                  formatted_date + "/" + region + \
                  "/s3/aws4_request, SignedHeaders=" + \
                                                     signed_headers_string + \
                                                     ", Signature=" + \
                                                                    signed_request
    return auth_header



# -*- coding: utf-8 -*-
# Minio Python Library for Amazon S3 compatible cloud storage, (C) 2015 Minio, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Helper functions
"""

import collections
import binascii
import hashlib
import re

from .compat import urlsplit, strtype, urlencode
from .error import InvalidBucketError, InvalidURLError

def get_region(hostname):
    """
    Get region based on hostname, defaults to "milkyway" if KeyError
    """
    region = {
        's3.amazonaws.com': 'us-east-1',
        's3-ap-northeast-1.amazonaws.com': 'ap-northeast-1',
        's3-ap-southeast-1.amazonaws.com': 'ap-southeast-1',
        's3-ap-southeast-2.amazonaws.com': 'ap-southeast-2',
        's3-eu-central-1.amazonaws.com': 'eu-central-1',
        's3-sa-east-1.amazonaws.com': 'sa-east-1',
        's3-external-1.amazonaws.com': 'us-east-1',
        's3-us-west-1.amazonaws.com': 'us-west-1',
        's3-us-west-2.amazonaws.com': 'us-west-2',
        's3.cn-north-1.amazonaws.com.cn': 'cn-north-1',
        's3-fips-us-gov-west-1.amazonaws.com': 'us-gov-west-1',
    }

    try:
        return region[hostname]
    except KeyError:
        return 'milkyway'

def get_target_url(url, bucket=None, key=None, query=None):
    """
    Construct target url
    """
    url_components = [url, '/']
    if key is not None:
        key = encode_object_key(key)

    if bucket is not None:
        url_components.append(bucket)
        if key is not None:
            url_components.append('/')
            url_components.append(key)

    if query is not None:
        ordered_query = collections.OrderedDict(sorted(query.items()))
        query_components = []
        for component_key in ordered_query:
            single_component = [component_key]
            if ordered_query[component_key] is not None:
                single_component.append('=')
                single_component.append(
                    urlencode(str(ordered_query[component_key])).replace('/', '%2F'))
            query_components.append(''.join(single_component))

        query_string = '&'.join(query_components)
        if query_string:
            url_components.append('?')
            url_components.append(query_string)

    return ''.join(url_components)

def is_valid_url(endpoint_url):
    """
    Verify the endpoint_url is valid.
    :type endpoint_url: string
    :param endpoint_url: An endpoint_url.  Must have at least a scheme
        and a hostname.
    :return: True if the endpoint url is valid. False otherwise.
    """
    if not isinstance(endpoint_url, strtype):
        raise TypeError('url')

    parts = urlsplit(endpoint_url)
    hostname = parts.hostname
    if hostname is None:
        raise InvalidURLError('url')
    if len(hostname) > 255:
        raise InvalidURLError('url')
    if hostname[-1] == ".":
        hostname = hostname[:-1]
    allowed = re.compile(
        "^((?!-)[A-Z\\d-]{1,63}(?<!-)\\.)*((?!-)[A-Z\\d-]{1,63}(?<!-))$",
        re.IGNORECASE)
    if not allowed.match(hostname):
        raise InvalidURLError('url')

def is_valid_bucket_name(bucket_name):
    """
    Check to see if the ``bucket_name`` complies with the
    restricted DNS naming conventions necessary to allow
    access via virtual-hosting style.

    Even though "." characters are perfectly valid in this DNS
    naming scheme, we are going to punt on any name containing a
    "." character because these will cause SSL cert validation
    problems if we try to use virtual-hosting style addressing.
    """
    validbucket = re.compile('[a-z0-9][a-z0-9\\-]*[a-z0-9]')
    if '.' in bucket_name:
        raise InvalidBucketError('bucket')
    if len(bucket_name) < 3 or len(bucket_name) > 63:
        # Wrong length
        raise InvalidBucketError('bucket')
    if len(bucket_name) == 1:
        if not bucket_name.isalnum():
            raise InvalidBucketError('bucket')
    match = validbucket.match(bucket_name)
    if match is None or match.end() != len(bucket_name):
        raise InvalidBucketError('bucket')

def is_non_empty_string(input_string):
    """
    validate if non empty string
    """
    if not isinstance(input_string, strtype):
        raise TypeError()
    if not input_string.strip():
        raise ValueError()

def encode_object_key(key):
    """
    url encode object key
    """
    is_non_empty_string(key)
    return urlencode(key)

def get_sha256(content):
    """
    calculate sha256 for given content
    """
    if len(content) == 0:
        content = b''
    hasher = hashlib.sha256()
    hasher.update(content)
    return hasher.digest()

def get_md5(content):
    """
    calculate md5 for given content
    """
    if len(content) == 0:
        content = b''
    hasher = hashlib.md5()
    hasher.update(content)
    return hasher.digest()

def encode_to_base64(content):
    """
    calculate base64 for given content
    """
    return binascii.b2a_base64(content).strip().decode('utf-8')

def encode_to_hex(content):
    """
    calculate hex for given content
    """
    return binascii.hexlify(content)

def calculate_part_size(length):
    """
    calculate optimal part size for multipart uploads
    """
    minimum_part_size = 1024 * 1024 * 5
    maximum_part_size = 1024 * 1024 * 1024 * 5
    proposed_part_size = length / 9999
    if proposed_part_size > maximum_part_size:
        return maximum_part_size
    return max(minimum_part_size, proposed_part_size)



# -*- coding: utf-8 -*-
# Minio Python Library for Amazon S3 compatible cloud storage, (C) 2015 Minio, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import io
import platform

import urllib3
import certifi

__author__ = "Minio, Inc."

from io import RawIOBase

from .__version__ import get_version
from .acl import is_valid_acl
from .compat import urlsplit, strtype
from .generators import (ListObjectsIterator, ListIncompleteUploads,
                         ListUploadParts, DataStreamer)
from .helpers import (get_target_url, is_non_empty_string, is_valid_url,
                      get_sha256, encode_to_base64, get_md5,
                      calculate_part_size, encode_to_hex,
                      is_valid_bucket_name, get_region)
from .parsers import (parse_list_buckets, parse_acl, parse_error,
                      parse_new_multipart_upload)
from .error import ResponseError
from .definitions import Object
from .signer import sign_v4
from .xml_requests import bucket_constraint, get_complete_multipart_upload

class Minio(object):
    def __init__(self, url, access_key=None, secret_key=None, certs=None):
        """
        Creates a new cloud storage client.

        Examples:

          client = Minio('https://play.minio.io:9000')
          client = Minio('https://s3.amazonaws.com', 'ACCESS_KEY', 'SECRET_KEY')

        :param url: A string of the URL of the cloud storage server.
        :param access_key: Access key to sign self._http.request with.
        :param secret_key: Secret key to sign self._http.request with.
        :param certs: Path to SSL certificates
        :return: Minio object
        """
        is_valid_url(url)

        url_components = urlsplit(url)
        self._location = url_components.netloc
        self._endpoint_url = url_components.geturl()
        self._access_key = access_key
        self._secret_key = secret_key
        self._user_agent = 'minio-py/' + get_version() + \
                           ' (' + platform.system() + '; ' + \
                           platform.machine() + ')'
        if certs is None:
            certs = certifi.where()

        self._http = urllib3.PoolManager(
            cert_reqs='CERT_REQUIRED',
            ca_certs=certs
        )

    # Client level
    def set_user_agent(self, name=None, version=None, comments=None):
        """
        Adds an entry to the list of user agents.

        Example:
            minio.add_user_agent('my_app', '1.0.0', ['ex', 'parrot'])
            # Results in my_app/1.0.0 (ex; parrot) appended to user agent

        :param name: user agent name
        :param version: user agent version
        :param comments: list of comments to include in comments section
        :return: None
        """
        if name is None or version is None:
            raise TypeError()
        if not isinstance(name, strtype) or \
           not isinstance(version, strtype):
            raise TypeError()
        if not name.strip() or not version.strip():
            raise ValueError()

        if comments is not None:
            joined_comments = '; '.join(comments)
            components = [' ', name, '/', version, ' (', joined_comments, ')']
            self._user_agent += ''.join(components)
        else:
            components = [' ', name, '/', version, ' ']
            self._user_agent += ''.join(components)

    # Bucket level
    def make_bucket(self, bucket, acl=None):
        """
        Make a new bucket on the server.

        Optionally include an ACL. Valid ACLs are as follows:

            Acl.public_read_write()
            Acl.public_read()
            Acl.authenticated_read()
            Acl.private()

        Examples:
            minio.make_bucket('foo')
            minio.make_bucket('foo', Acl.public_read())

        :param bucket: Bucket to create on server
        :param acl: Canned ACL to use. Default is Acl.private()
        :return:
        """
        is_valid_bucket_name(bucket)
        if acl is not None:
            is_valid_acl(acl)

        method = 'PUT'
        url = get_target_url(self._endpoint_url, bucket=bucket)
        headers = {}

        if acl is not None:
            headers['x-amz-acl'] = acl

        region = get_region(self._location)

        content = ''
        if not (region == 'us-east-1' or region == 'milkyway'):
            content = bucket_constraint(region)
            headers['Content-Length'] = str(len(content))

        content_sha256 = get_sha256(content)
        if content.strip():
            content_md5 = encode_to_base64(get_md5(content))
            headers['Content-MD5'] = content_md5


        headers = sign_v4(method=method, url=url, headers=headers,
                          access_key=self._access_key,
                          secret_key=self._secret_key,
                          content_hash=content_sha256)

        response = self._http.urlopen(method, url, body=content,
                                      headers=headers)

        if response.status != 200:
            parse_error(response, bucket)

    def list_buckets(self):
        """
        List all buckets owned by the user.


        Example:
            bucket_list = minio.list_buckets()
            for bucket in bucket_list:
                print bucket.name,bucket.created_date

        :return: A list of buckets owned by the current user.
        """
        url = get_target_url(self._endpoint_url)
        method = 'GET'
        headers = {}


        headers = sign_v4(method=method, url=url, headers=headers,
                          access_key=self._access_key,
                          secret_key=self._secret_key)

        response = self._http.request(method, url, headers=headers,
                                      redirect=False)

        if response.status != 200:
            try:
                parse_error(response)
            except ResponseError as err:
                if err.code == 'Redirect':
                    err.code = 'AccessDeniedException'
                raise err
        return parse_list_buckets(response.data)

    def bucket_exists(self, bucket):
        """
        Check if the bucket exists and if the user has access to it.

        :param bucket: To test the existence and user access.
        :return: True on success. Otherwise, returns False
        """
        is_valid_bucket_name(bucket)

        method = 'HEAD'
        url = get_target_url(self._endpoint_url, bucket=bucket)
        headers = {}


        headers = sign_v4(method=method, url=url, headers=headers,
                          access_key=self._access_key,
                          secret_key=self._secret_key)

        response = self._http.request(method, url, headers=headers)

        if response.status != 200:
            if response.status == "404":
                return False
            parse_error(response, bucket)

        return True

    def remove_bucket(self, bucket):
        """
        Remove a bucket.

        :param bucket: Bucket to remove
        :return: None
        """
        is_valid_bucket_name(bucket)

        method = 'DELETE'
        url = get_target_url(self._endpoint_url, bucket=bucket)
        headers = {}


        headers = sign_v4(method=method, url=url, headers=headers,
                          access_key=self._access_key,
                          secret_key=self._secret_key)

        response = self._http.request(method, url, headers=headers)

        if response.status != 204:
            parse_error(response, bucket)

    def get_bucket_acl(self, bucket):
        """
        Get a bucket's canned ACL, if any.

        Example:
            canned_acl = minio.get_bucket_acl('foo')
            if canned_acl == Acl.private():
                # do something

        :param bucket: Bucket to check canned ACL of.
        :return: A string representing canned ACL on the bucket.
        """
        is_valid_bucket_name(bucket)

        method = 'GET'
        url = get_target_url(self._endpoint_url, bucket=bucket,
                             query={"acl": None})
        headers = {}

        headers = sign_v4(method=method, url=url, headers=headers,
                          access_key=self._access_key,
                          secret_key=self._secret_key)

        response = self._http.request(method, url, headers=headers)

        if response.status != 200:
            parse_error(response, bucket)

        return parse_acl(response.data)

    def set_bucket_acl(self, bucket, acl):
        """
        Set a bucket's canned acl

        Valid ACLs include:
            Acl.public_read_write()
            Acl.public_read()
            Acl.authenticated_read()
            Acl.private()

        Example:
            canned_acl = minio.get_bucket_acl('foo')
            if canned_acl == Acl.private():
                # do something

        :param bucket: Bucket to set
        :param acl: ACL to set
        :return: None
        """
        is_valid_bucket_name(bucket)
        is_valid_acl(acl)

        method = 'PUT'
        url = get_target_url(self._endpoint_url, bucket=bucket,
                             query={"acl": None})

        headers = {
            'x-amz-acl': acl,
        }

        headers = sign_v4(method=method, url=url, headers=headers,
                          access_key=self._access_key,
                          secret_key=self._secret_key)

        response = self._http.urlopen(method, url, headers=headers)

        if response.status != 200:
            parse_error(response, bucket)

    def drop_all_incomplete_uploads(self, bucket):
        """
        Drop all incomplete uploads in a bucket.

        :param bucket: Bucket to drop all incomplete uploads.
        :return: None
        """
        # check bucket
        is_valid_bucket_name(bucket)

        uploads = ListIncompleteUploads(self._http, self._endpoint_url,
                                        bucket, None,
                                        access_key=self._access_key,
                                        secret_key=self._secret_key)

        for upload in uploads:
            self._drop_incomplete_upload(bucket, upload.key, upload.upload_id)

    def get_object(self, bucket, key):
        """
        Retrieves an object from a bucket.

        Examples:
            my_partial_object = minio.get_partial_object('foo', 'bar')

        :param bucket: Bucket to retrieve object from
        :param key: Key to retrieve
        :return: An iterable containing a byte stream of the data.
        """
        return self.get_partial_object(bucket, key)

    # Object Level
    def get_partial_object(self, bucket, key, offset=0, length=0):
        """
        Retrieves an object from a bucket.

        Optionally takes an offset and length of data to retrieve.

        Examples:
            my_partial_object = minio.get_partial_object('foo', 'bar', 2, 4)

        :param bucket: Bucket to retrieve object from
        :param key: Key to retrieve
        :param offset: Optional offset to retrieve bytes from. Must be >= 0
        :param length: Optional number of bytes to retrieve. Must be > 0
        :return: An iterable containing a byte stream of the data.
        """
        is_valid_bucket_name(bucket)
        is_non_empty_string(key)

        request_range = ''
        if offset is not 0 and length is not 0:
            request_range = str(offset) + "-" + str(offset + length - 1)
        if offset is not 0 and length is 0:
            request_range = str(offset) + "-"
        if offset is 0 and length is not 0:
            request_range = "0-" + str(length - 1)

        method = 'GET'
        url = get_target_url(self._endpoint_url, bucket=bucket, key=key)
        headers = {}

        if request_range:
            headers['Range'] = 'bytes=' + request_range

        headers = sign_v4(method=method, url=url, headers=headers,
                          access_key=self._access_key,
                          secret_key=self._secret_key)

        response = self._http.urlopen(method, url, headers=headers,
                                      preload_content=False)

        if response.status != 206 and response.status != 200:
            parse_error(response, bucket+"/"+key)

        return DataStreamer(response)

    def put_object(self, bucket, key, length, data,
                   content_type="application/octet-stream"):
        """
        Add a new object to the cloud storage server.

        Data can either be a string, byte array, or reader (e.g. open('foo'))

        Examples:
            minio.put('foo', 'bar', 11, 'hello world')

            minio.put('foo', 'bar', 11, b'hello world', 'text/plain')

            with open('hello.txt', 'rb') as data:
                minio.put('foo', 'bar', 11, b'hello world', 'text/plain')

        :param bucket: Bucket of new object.
        :param key: Key of new object.
        :param length: Total length of object.
        :param data: Contents to upload.
        :param content_type: mime type of object as a string.
        :return: None
        """
        is_valid_bucket_name(bucket)
        is_non_empty_string(key)

        if length is 0:
            raise ValueError('length')

        if length <= 5 * 1024 * 1024:
            # reference 'file' for python 2.7 compatibility, RawIOBase for 3.X
            if type(data).__name__ == 'file' or \
               isinstance(data, io.BufferedReader):
                data = data.read(length)
            if isinstance(data, io.TextIOWrapper):
                data = data.read(length).encode('utf-8')
            if sys.version_info >= (3, 0) and isinstance(data, strtype):
                data = data.encode('utf-8')
            return self._do_put_object(bucket, key, length, data, content_type)
        self._stream_put_object(bucket, key, length, data, content_type)

    def list_objects(self, bucket, prefix=None, recursive=False):
        """
        List objects in the given bucket.

        Examples:
            objects = minio.list_objects('foo')
            for current_object in objects:
                print current_object
            # hello
            # hello/
            # hello/
            # world/

            objects = minio.list_objects('foo', prefix='hello/')
            for current_object in objects:
                print current_object
            # hello/world/

            objects = minio.list_objects('foo', recursive=True)
            for current_object in objects:
                print current_object
            # hello/world/1
            # world/world/2
            # ...

            objects = minio.list_objects('foo', prefix='hello/',
                                         recursive=True)
            for current_object in objects:
                print current_object
            # hello/world/1
            # hello/world/2

        :param bucket: Bucket to list objects from
        :param prefix: String specifying objects returned must begin with
        :param recursive: If yes, returns all objects for a specified prefix
        :return: An iterator of objects in alphabetical order.
        """
        is_valid_bucket_name(bucket)
        return ListObjectsIterator(self._http, self._endpoint_url, bucket,
                                   prefix, recursive, self._access_key,
                                   self._secret_key)

    def stat_object(self, bucket, key):
        """
        Check if an object exists.

        :param bucket: Bucket of object.
        :param key: Key of object
        :return: Object metadata if object exists
        """
        is_valid_bucket_name(bucket)
        is_non_empty_string(key)

        method = 'HEAD'
        url = get_target_url(self._endpoint_url, bucket=bucket, key=key)
        headers = {}


        headers = sign_v4(method=method, url=url, headers=headers,
                          access_key=self._access_key,
                          secret_key=self._secret_key)

        response = self._http.request(method, url, headers=headers)

        if response.status != 200:
            parse_error(response, bucket+"/"+key)

        content_type = response.headers['content-type']
        etag = response.headers['etag'].replace('"', '')
        size = response.headers['content-length']
        last_modified = response.headers['last-modified']

        return Object(bucket, key, content_type=content_type,
                      last_modified=last_modified, etag=etag, size=size)

    def remove_object(self, bucket, key):
        """
        Remove an object from the bucket.

        :param bucket: Bucket of object to remove
        :param key: Key of object to remove
        :return: None
        """
        is_valid_bucket_name(bucket)
        is_non_empty_string(key)

        method = 'DELETE'
        url = get_target_url(self._endpoint_url, bucket=bucket, key=key)
        headers = {}

        headers = sign_v4(method=method, url=url, headers=headers,
                          access_key=self._access_key,
                          secret_key=self._secret_key)

        response = self._http.urlopen(method, url, headers=headers)

        if response.status != 204:
            parse_error(response, bucket+"/"+key)

    def drop_incomplete_upload(self, bucket, key):
        """
        Drops all in complete uploads for a given bucket and key.

        :param bucket: Bucket to drop incomplete uploads
        :param key: Key of object to drop incomplete uploads of
        :return: None
        """
        is_valid_bucket_name(bucket)
        is_non_empty_string(key)

        # check key
        uploads = ListIncompleteUploads(self._http, self._endpoint_url,
                                        bucket, key,
                                        access_key=self._access_key,
                                        secret_key=self._secret_key)
        for upload in uploads:
            self._drop_incomplete_upload(bucket, upload.key, upload.upload_id)

    # helper functions

    def _do_put_object(self, bucket, key, length, data,
                       content_type='application/octet-stream',
                       upload_id='', part_number=0):
        method = 'PUT'

        if len(data) != length:
            raise DataSizeMismatchError()

        if upload_id.strip() and part_number is not 0:
            url = get_target_url(self._endpoint_url, bucket=bucket, key=key,
                                 query={'uploadId': upload_id,
                                        'partNumber': part_number})
        else:
            url = get_target_url(self._endpoint_url, bucket=bucket, key=key)

        content_sha256 = get_sha256(data)
        content_md5 = encode_to_base64(get_md5(data))

        headers = {
            'Content-Length': length,
            'Content-Type': content_type,
            'Content-MD5': content_md5
        }

        headers = sign_v4(method=method, url=url, headers=headers,
                          access_key=self._access_key,
                          secret_key=self._secret_key,
                          content_hash=content_sha256)

        data = io.BytesIO(data)
        response = self._http.urlopen(method, url, headers=headers, body=data)

        if response.status != 200:
            parse_error(response, bucket+"/"+key)

        return response.headers['etag'].replace('"', '')

    def _stream_put_object(self, bucket, key, length, data, content_type):
        if type(data).__name__ != 'file':
            if not isinstance(data, io.BufferedReader):
                if not isinstance(data, RawIOBase):
                    if sys.version_info >= (3, 0):
                        if isinstance(data, strtype):
                            data = data.encode('utf-8')
                    data = io.BytesIO(data)
                data = io.BufferedReader(data)

        part_size = calculate_part_size(length)

        current_uploads = ListIncompleteUploads(self._http,
                                                self._endpoint_url,
                                                bucket,
                                                key,
                                                access_key=self._access_key,
                                                secret_key=self._secret_key)

        upload_id = None
        for upload in current_uploads:
            upload_id = upload.upload_id
        uploaded_parts = {}
        if upload_id is not None:
            part_iter = ListUploadParts(self._http, self._endpoint_url,
                                        bucket, key, upload_id,
                                        access_key=self._access_key,
                                        secret_key=self._secret_key)
            for part in part_iter:
                uploaded_parts[part.part_number] = part
        else:
            upload_id = self._new_multipart_upload(bucket, key, content_type)
        total_uploaded = 0
        current_part_number = 1
        etags = []
        while total_uploaded < length:
            current_data = data.read(part_size)
            if len(current_data) == 0:
                break
            current_data_md5 = encode_to_hex(get_md5(current_data))
            previously_uploaded_part = None
            if current_part_number in uploaded_parts:
                previously_uploaded_part = uploaded_parts[current_part_number]
            if previously_uploaded_part is None or \
               previously_uploaded_part.etag != current_data_md5:
                etag = self._do_put_object(bucket=bucket, key=key,
                                           length=len(current_data),
                                           data=current_data,
                                           content_type=content_type,
                                           upload_id=upload_id,
                                           part_number=current_part_number)
            else:
                etag = previously_uploaded_part.etag
            etags.append(etag)
            total_uploaded += len(current_data)
            current_part_number += 1
        if total_uploaded != length:
            raise DataSizeMismatchError()
        self._complete_multipart_upload(bucket, key, upload_id, etags)

    def _drop_incomplete_upload(self, bucket, key, upload_id):
        method = 'DELETE'
        query = {
            'uploadId': upload_id
        }
        url = get_target_url(self._endpoint_url, bucket=bucket, key=key, query=query)
        headers = {}

        headers = sign_v4(method=method, url=url, headers=headers,
                          access_key=self._access_key,
                          secret_key=self._secret_key)

        response = self._http.request(method, url, headers=headers)

        if response.status != 204:
            parse_error(response, bucket+"/"+key)

    def _new_multipart_upload(self, bucket, key, content_type):
        method = 'POST'
        query = {
            'uploads': None
        }

        url = get_target_url(self._endpoint_url, bucket=bucket,
                             key=key, query=query)

        headers = {
            'Content-Type': content_type
        }


        headers = sign_v4(method=method, url=url, headers=headers,
                          access_key=self._access_key,
                          secret_key=self._secret_key)

        response = self._http.urlopen(method, url, headers=headers, body=None)

        if response.status != 200:
            parse_error(response, bucket+"/"+key)

        return parse_new_multipart_upload(response.data)

    def _complete_multipart_upload(self, bucket, key, upload_id, etags):
        method = 'POST'
        query = {
            'uploadId': upload_id
        }
        url = get_target_url(self._endpoint_url, bucket=bucket,
                             key=key, query=query)
        headers = {}

        data = get_complete_multipart_upload(etags)
        data_sha256 = get_sha256(data)
        data_md5 = encode_to_base64(get_md5(data))

        headers['Content-Length'] = len(data)
        headers['Content-Type'] = 'application/xml'
        headers['Content-MD5'] = data_md5


        headers = sign_v4(method=method, url=url, headers=headers,
                          access_key=self._access_key,
                          secret_key=self._secret_key, content_hash=data_sha256)

        response = self._http.urlopen(method, url, headers=headers, body=data)

        if response.status != 200:
            parse_error(response, bucket+"/"+key)

class DataSizeMismatchError(BaseException):
    pass



# -*- coding: utf-8 -*-
# Minio Python Library for Amazon S3 compatible cloud storage, (C) 2015 Minio, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# version_info should conform to PEP 386
# (major, minor, micro, alpha/beta/rc/final, #)
# (1, 1, 2, 'alpha', 0) => "1.1.2.dev"
# (1, 2, 0, 'beta', 2) => "1.2b2"
version_info = (0, 2, 4, 'final', 0)

def get_version():
    " Returns a PEP 386-compliant version number from version_info. "
    assert len(version_info) == 5
    assert version_info[3] in ('alpha', 'beta', 'rc', 'final')

    parts = 2 if version_info[2] == 0 else 3
    main = '.'.join(map(str, version_info[:parts]))

    sub = ''
    if version_info[3] == 'alpha' and version_info[4] == 0:
        sub = '.dev0'
    elif version_info[3] != 'final':
        mapping = {'alpha': 'a', 'beta': 'b', 'rc': 'c'}
        sub = mapping[version_info[3]] + str(version_info[4])

    return str(main + sub)



# -*- coding: utf-8 -*-
# Minio Python Library for Amazon S3 compatible cloud storage, (C) 2015 Minio, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .helpers import get_target_url
from .parsers import (parse_list_objects, parse_error,
                      parse_incomplete_uploads, parse_uploaded_parts)
from .signer import sign_v4

class ListObjectsIterator(object):
    def __init__(self, client, url, bucket, prefix,
                 recursive, access_key, secret_key):
        self._http = client
        self._endpoint_url = url
        self._bucket = bucket
        self._prefix = prefix
        self._recursive = recursive
        self._results = []
        self._complete = False
        self._access_key = access_key
        self._secret_key = secret_key
        self._is_truncated = True
        self._marker = None

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        # if complete, end iteration
        if self._complete:
            raise StopIteration
        # if not truncated and we've emitted everything, end iteration
        if len(self._results) == 0 and self._is_truncated is False:
            self._complete = True
            raise StopIteration
        # perform another fetch
        if len(self._results) == 0:
            self._results, self._is_truncated, self._marker = self._fetch()
        # if fetch results in no elements, end iteration
        if len(self._results) == 0:
            self._complete = True
            raise StopIteration
        # return result
        return self._results.pop(0)

    def _fetch(self):
        query = {}
        query['max-keys'] = 1000
        if self._prefix is not None:
            query['prefix'] = self._prefix
        if not self._recursive:
            query['delimiter'] = '/'
        if self._marker is not None:
            query['marker'] = self._marker

        url = get_target_url(self._endpoint_url, bucket=self._bucket, query=query)

        method = 'GET'
        headers = {}

        headers = sign_v4(method=method, url=url, headers=headers,
                          access_key=self._access_key,
                          secret_key=self._secret_key)

        response = self._http.request(method, url, headers=headers)

        if response.status != 200:
            parse_error(response, self._bucket)

        return parse_list_objects(response.data, bucket=self._bucket)


class ListIncompleteUploads(object):
    def __init__(self, client, url, bucket, key=None,
                 access_key=None, secret_key=None):
        # from user
        self._http = client
        self._endpoint_url = url
        self._bucket = bucket
        self._key = key
        self._access_key = access_key
        self._secret_key = secret_key

        # internal variables
        self._results = []
        self._complete = False
        self._is_truncated = True
        self._key_marker = None
        self._upload_id_marker = None

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        # if complete, end iteration
        if self._complete:
            raise StopIteration
        # if not truncated and we've emitted everything, end iteration
        if len(self._results) == 0 and self._is_truncated is False:
            self._complete = True
            raise StopIteration
        # perform another fetch
        if len(self._results) == 0:
            self._results, self._is_truncated, self._key_marker, \
                self._upload_id_marker = self._fetch()
        # if fetch results in no elements, end iteration
        if len(self._results) == 0:
            self._complete = True
            raise StopIteration
        # return result
        potential_result = self._results.pop(0)
        if self._key is None:
            return potential_result
        if potential_result.key == self._key:
            return potential_result
        self._complete = True
        raise StopIteration

    def _fetch(self):
        query = {
            'uploads': None
        }
        query['max-uploads'] = 1000
        if self._key is not None:
            query['prefix'] = self._key
        if self._key_marker is not None:
            query['key-marker'] = self._key_marker
        if self._upload_id_marker is not None:
            query['upload-id-marker'] = self._upload_id_marker

        url = get_target_url(self._endpoint_url, bucket=self._bucket, query=query)

        method = 'GET'
        headers = {}

        headers = sign_v4(method=method, url=url, headers=headers,
                          access_key=self._access_key,
                          secret_key=self._secret_key)

        response = self._http.request(method, url, headers=headers)

        if response.status != 200:
            parse_error(response, self._bucket)

        return parse_incomplete_uploads(response.data, bucket=self._bucket)


class ListUploadParts(object):
    def __init__(self, client, url, bucket, key, upload_id,
                 access_key=None, secret_key=None):
        # from user
        self._http = client
        self._endpoint_url = url
        self._bucket = bucket
        self._key = key
        self._upload_id = upload_id
        self._access_key = access_key
        self._secret_key = secret_key

        # internal variables
        self._results = []
        self._complete = False
        self._is_truncated = True
        self._part_marker = None

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        # if complete, end iteration
        if self._complete:
            raise StopIteration
        # if not truncated and we've emitted everything, end iteration
        if len(self._results) == 0 and self._is_truncated is False:
            self._complete = True
            raise StopIteration
        # perform another fetch
        if len(self._results) == 0:
            self._results, self._is_truncated, self._part_marker = self._fetch()
        # if fetch results in no elements, end iteration
        if len(self._results) == 0:
            self._complete = True
            raise StopIteration
        # return result
        potential_result = self._results.pop(0)
        if self._key is None:
            return potential_result
        if potential_result.key == self._key:
            return potential_result
        self._complete = True
        raise StopIteration

    def _fetch(self):
        query = {
            'uploadId': self._upload_id
        }
        query['max-parts'] = 1000
        if self._part_marker is not None:
            query['part-number-marker'] = self._part_marker

        url = get_target_url(self._endpoint_url, bucket=self._bucket,
                             key=self._key, query=query)

        method = 'GET'
        headers = {}

        headers = sign_v4(method=method, url=url, headers=headers,
                          access_key=self._access_key,
                          secret_key=self._secret_key)

        response = self._http.request(method, url, headers=headers)

        if response.status != 200:
            parse_error(response, self._bucket+"/"+self._key)

        return parse_uploaded_parts(response.data, bucket=self._bucket,
                                    key=self._key, upload_id=self._upload_id)


class DataStreamer(object):
    def __init__(self, response):
        self._response = response
        self._stream = iter(response.stream())

        self._length = int(response.headers['content-length'])
        self._total_read = 0
        self._is_complete = False

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        if self._is_complete:
            raise StopIteration
        data = self._response.read(1024)
        self._total_read += len(data)
        if self._total_read == self._length:
            self._is_complete = True
        return data

    def _release(self):
        if not self._is_complete:
            self._response.release_conn()
            self._is_complete = True
        raise StopIteration



# -*- coding: utf-8 -*-
# Minio Python Library for Amazon S3 compatible cloud storage, (C) 2015 Minio, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class InvalidURLError(Exception):
    def __init__(self, message, **kwargs):
        self.message = message
        super(InvalidURLError, self).__init__(**kwargs)

    def __str__(self):
        string_format = 'InvalidURLError: message: {0}'
        return string_format.format(self.message)

class InvalidBucketError(Exception):
    def __init__(self, message, **kwargs):
        self.message = message
        super(InvalidBucketError, self).__init__(**kwargs)

    def __str__(self):
        string_format = 'InvalidBucketError: message: {0}'
        return string_format.format(self.message)

class InvalidArgumentError(Exception):
    def __init__(self, message, **kwargs):
        self.message = message
        super(InvalidArgumentError, self).__init__(**kwargs)

    def __str__(self):
        string_format = 'InvalidArgumentError: message: {0}'
        return string_format.format(self.message)


class ResponseError(Exception):
    def __init__(self, code, message, request_id, host_id, resource, xml=None,
                 **kwargs):
        super(ResponseError, self).__init__(**kwargs)
        self.code = code
        self.message = message
        self.request_id = request_id
        self.host_id = host_id
        self.resource = resource
        self.xml = xml

    def __str__(self):
        return 'ResponseError: code: {0}, message: {1}, request_id: {2},' \
            'host_id: {3}, resource: {4}, xml: {5}'.format(self.code,
                                                           self.message,
                                                           self.request_id,
                                                           self.host_id,
                                                           self.resource,
                                                           self.xml)



# -*- coding: utf-8 -*-
# Minio Python Library for Amazon S3 compatible cloud storage, (C) 2015 Minio, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

try:
    from urllib.parse import urlsplit
except ImportError:  # python 2
    from urlparse import urlsplit

try:
    from urllib.request import pathname2url as urlencode
except ImportError:  # python 2
    from urllib import pathname2url as urlencode

try:
    from urllib.request import url2pathname as urldecode
except ImportError:  # python 2
    from urllib import url2pathname as urldecode


strtype = None
if sys.version_info < (3, 0):
    strtype = basestring
else:
    strtype = str
